from gbd_tool.gbd_api import GBD

from graph_tool.all import *
from pysat.formula import CNF

import itertools


def gen_vig(path):
    g = Graph(directed=False)
    cnf = CNF()
    cnf.from_file(path, compressed_with='lzma')
    print("NVars: {}".format(cnf.nv))
    g.add_vertex(cnf.nv + 1)
    w = g.new_edge_property("float")
    for clause in cnf:
        inc = 1.0 / pow(2, len(clause))
        for tup in itertools.combinations(clause, 2):
            e = g.edge(g.vertex(abs(tup[0])), g.vertex(abs(tup[1])), add_missing=True)
            w[e] = w[e] + inc
    return (g, w)


def vig(api: GBD, path, proof):
    (g, w) = gen_vig(path)
    pos = sfdp_layout(g, eweight=w)
    graph_draw(g, pos, output_size=(1000, 1000), vertex_color=[1,1,1,0],
           vertex_size=1, edge_pen_width=1.2, output="graph.pdf")

def read_by_tokens(fileobj):
    for line in fileobj:
        for token in line.split():
            yield token

from gi.repository import Gtk, Gdk, GdkPixbuf, GObject, GLib

def animate_proof(api: GBD, path, proof):
    (g, w) = gen_vig(path)
    pos = sfdp_layout(g, eweight=w)
    win = GraphWindow(g, pos, geometry=(800, 600))

    handle = open(proof)        
    def update(handle):
        delete = False
        clause = []
        sfdp_layout(g, pos=pos, K=0.5 , init_step=0.005, max_iter=1)
        for token in read_by_tokens(handle):
            if token == "d":
                delete = True
            elif token == "0":
                inc = 1.0 / pow(2, len(clause))
                if delete:
                    for tup in itertools.combinations(clause, 2):
                        e = g.edge(g.vertex(abs(tup[0])), g.vertex(abs(tup[1])))
                        w[e] = w[e] - inc
                else:
                    for tup in itertools.combinations(clause, 2):
                        e = g.edge(g.vertex(abs(tup[0])), g.vertex(abs(tup[1])))
                        w[e] = w[e] + inc
                print(clause)
                win.graph.regenerate_surface()
                win.graph.queue_draw()
                return True
            else:
                clause = clause + [int(token)]
        return False
        
    cid = GLib.idle_add(update, handle)
   
    # We will give the user the ability to stop the program by closing the window.

    win.connect("delete_event", Gtk.main_quit)


    # Actually show the window, and start the main loop.

    win.show_all()

    Gtk.main()




# GBD Benchmark Database (GBD)
# Copyright (C) 2021 Markus Iser, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sqlite3
import os
import csv

from dataclasses import dataclass

import gbd_tool.config as config
from gbd_tool.gbd_hash import HASH_VERSION
from gbd_tool.util import eprint, prepend_context, context_from_name, make_alnum_ul


VERSION = 1


class SchemaException(Exception):
    pass


@dataclass
class FeatureInfo:
    name: str = None
    database: str = None
    context: str = None
    table: str = None
    column: str = None
    default: str = None
    virtual: bool = False


class Schema:
    IN_MEMORY_DB = "file::memory:?cache=shared"

    def __init__(self, path):
        self.contexts = []
        self.tables = []
        self.views = []
        self.features = []
        if self.is_database(path):
            self.dbname = make_alnum_ul(os.path.basename(path))
            self.path = path
            self.connection = sqlite3.connect(path)
            self.csv = False
            self.check_version(VERSION, HASH_VERSION)
            self.schema_from_database()
        else:
            self.dbname = "main"
            self.path = path
            self.connection = sqlite3.connect(self.IN_MEMORY_DB, uri=True)
            self.csv = True
            self.schema_from_csv()

    @classmethod
    def is_database(cls, path):
        if not os.path.isfile(path): 
            Schema.create(path, VERSION, HASH_VERSION)
            return True
        sz = os.path.getsize(path)
        if sz == 0: return True  # new sqlite3 files can be empty
        if sz < 100: return False  # sqlite header is 100 bytes
        with open(path, 'rb') as fd: header = fd.read(100)  # validate header
        return (header[:16] == b'SQLite format 3\x00')

    @classmethod
    def create(cls, path, version, hash_version):
        eprint("Warning: Creating Database {}".format(path))
        try:
            con = sqlite3.connect(path)
            con.cursor().execute("CREATE TABLE __version (version INT, hash_version INT)")
            con.cursor().execute("INSERT INTO __version (version, hash_version) VALUES ({}, {})".format(version, hash_version))
            con.commit()
            con.close()
        except Exception as e:
            raise SchemaException(str(e))

    def check_version(self, version, hash_version):
        try:
            __version = self.connection.execute("SELECT version, hash_version FROM __version").fetchall()
            if __version[0][0] != version:
                raise SchemaException("WARNING: Database schema version is {} but tool supported schema version is {}".format(__version[0][0], version))
            if __version[0][1] != hash_version:
                raise SchemaException("WARNING: Database hash version is {} but tool supported hash version is {}".format(__version[0][1], hash_version))
        except Exception as e:
            raise SchemaException(str(e))

    def absorb(self, schema):
        if not self.csv or not schema.csv:
            raise SchemaException("Internal Error: Attempt to merge non-virtual schemata")
        self.contexts.extend(schema.contexts)
        self.tables.extend(schema.tables)
        self.views.extend(schema.views)
        self.features.extend(schema.features)


    # Import CSV to IN_MEMORY_DB, create according schema info
    def schema_from_csv(self):
        filename = os.path.basename(self.path)
        table = make_alnum_ul(filename)
        context = context_from_name(filename)
        self.contexts.append(context)
        self.views.append(table)
        try:
            with open(self.path) as csvfile:
                csvreader = csv.DictReader(csvfile)
                if not "hash" in csvreader.fieldnames:
                    raise SchemaException("Column 'hash' not found in {}".format(csvfile))
                for title in csvreader.fieldnames:
                    self.features.append(FeatureInfo(prepend_context(title, context), self.dbname, context, table, title, None, True))
                self.connection.execute('CREATE TABLE IF NOT EXISTS {} ({})'.format(table, ", ".join(csvreader.fieldnames)))
                for row in csvreader:
                    self.connection.execute("INSERT INTO {} VALUES ('{}')".format(table, "', '".join(row.values())))
            self.connection.commit()
        except Exception as e:
            raise SchemaException(str(e))

    # Create schema info for sqlite database
    def schema_from_database(self):
        sql_tables="""SELECT tbl_name, type FROM sqlite_master WHERE type IN ('table', 'view') 
                        AND NOT tbl_name LIKE 'sqlite$_%' ESCAPE '$' AND NOT tbl_name LIKE '$_$_%' ESCAPE '$'"""
        tables = self.connection.execute(sql_tables).fetchall()
        for (table, type) in tables:
            context = context_from_name(table)
            if not context in self.contexts:
                self.contexts.append(context)
            if type == "view":
                self.views.append(table)
            else:
                self.tables.append(table)
            columns = self.connection.execute("PRAGMA table_info({})".format(table)).fetchall()
            names = ('index', 'name', 'type', 'notnull', 'default_value', 'pk')
            for record in columns:
                col = dict(zip(names, record))
                fname = table if col['name'] == "value" else col['name']
                if fname == "hash":
                    fname = prepend_context("hash", context)
                self.features.append(FeatureInfo(fname, self.dbname, context, table, col['name'], col['default_value'], type == "view"))
        for context in self.contexts:
            self.create_main_feature_table(context)


    def create_main_feature_table(self, context):
        self.valid_context_or_raise(context)
        main_table = prepend_context("features", context)
        if not main_table in self.tables:
            self.connection.execute("CREATE TABLE IF NOT EXISTS {} (hash UNIQUE NOT NULL)".format(main_table))
            for table in filter(lambda t: context == context_from_name(t), self.tables):
                self.connection.execute("INSERT OR IGNORE INTO {} (hash) SELECT DISTINCT(hash) FROM {}".format(main_table, table))
                self.connection.execute("""CREATE TRIGGER IF NOT EXISTS {}_dval AFTER INSERT ON {} 
                                            BEGIN INSERT OR IGNORE INTO {} (hash) VALUES (NEW.hash); END""".format(table, table, main_table))
                self.connection.commit()
            self.tables.append(main_table)
            if not context in self.contexts:
                self.contexts.append(context)
            self.features.append(FeatureInfo(prepend_context("hash", context), self.dbname, context, main_table, "hash", None, False))

    def create_context_translator_table(self, src, dst):
        self.valid_context_or_raise(src)
        self.valid_context_or_raise(dst)
        translator = "{}_to_{}".format(src, dst)
        if not translator in self.tables:
            self.connection.execute("CREATE TABLE IF NOT EXISTS {} (hash, value)".format(translator))
            self.connection.commit()
        self.tables.append(translator)
        if not src in self.contexts:
            self.contexts.append(src)
        self.features.append(FeatureInfo(prepend_context("hash", src), self.dbname, src, translator, "hash", None, False))


    @classmethod
    def is_main_hash_column(cls, info: FeatureInfo):
        is_main = prepend_context("features", info.context) == info.table
        is_hash = info.column == "hash"
        return is_main and is_hash

    @classmethod
    def valid_context_or_raise(cls, name):
        if not name in config.contexts():
            raise SchemaException("Unknown Context: " + name)

    @classmethod
    def valid_feature_or_raise(cls, name):
        if len(name) < 2:
            raise SchemaException("Feature name '{}' is to short.".format(name))
        gbd_keywords = [ 'hash', 'value', 'local', 'filename', 'features' ]
        if name.lower() in gbd_keywords or name.startswith("__"):
            raise SchemaException("Feature name '{}' is reserved.".format(name))
        sqlite_keywords = ['abort', 'action', 'add', 'after', 'all', 'alter', 'always', 'analyze', 'and', 'as', 'asc', 'attach', 'autoincrement', 
            'before', 'begin', 'between', 'by', 'cascade', 'case', 'cast', 'check', 'collate', 'column', 'commit', 'conflict', 'constraint', 
            'create', 'cross', 'current', 'current_date', 'current_time', 'current_timestamp', 'database', 'default', 'deferrable', 'deferred', 
            'delete', 'desc', 'detach', 'distinct', 'do', 'drop', 'each', 'else', 'end', 'escape', 'except', 'exclude', 'exclusive', 'exists', 
            'explain', 'fail', 'filter', 'first', 'following', 'for', 'foreign', 'from', 'full', 'generated', 'glob', 'group', 'groups', 
            'having', 'if', 'ignore', 'immediate', 'in', 'index', 'indexed', 'initially', 'inner', 'insert', 'instead', 'intersect', 'into', 'is', 'isnull', 
            'join', 'key', 'last', 'left', 'like', 'limit', 'match', 'materialized', 'natural', 'no', 'not', 'nothing', 'notnull', 'null', 'nulls', 
            'of', 'offset', 'on', 'or', 'order', 'others', 'outer', 'over', 'partition', 'plan', 'pragma', 'preceding', 'primary', 'query', 
            'raise', 'range', 'recursive', 'references', 'regexp', 'reindex', 'release', 'rename', 'replace', 'restrict', 'returning', 'right', 'rollback', 
            'row', 'rows', 'savepoint', 'select', 'set', 'table', 'temp', 'temporary', 'then', 'ties', 'to', 'transaction', 'trigger', 'unbounded', 'union', 
            'unique', 'update', 'using', 'vacuum', 'values', 'view', 'virtual', 'when', 'where', 'window', 'with', 'without']
        if name.lower() in sqlite_keywords or name.startswith("sqlite_"):
            raise SchemaException("Feature name '{}' is reserved by sqlite.".format(name))



# GBD Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



contexts_ = {
    "cnf" : [".cnf", ".cnf.gz", ".cnf.lzma", ".cnf.xz", ".cnf.bz2"],
    "kis" : [".kis", ".kis.gz", ".kis.lzma", ".kis.xz", ".kis.bz2"]
}

def suffix_list(context):
    return contexts_[context]

def contexts():
    return contexts_.keys()


# GBD Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Luca Springer, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sqlite3
import multiprocessing
import tatsu
import os

from contextlib import ExitStack

from gbd_tool.query_builder import GBDQuery
from gbd_tool.db import Database
from gbd_tool.util import eprint


class GBDException(Exception):
    pass


class GBD:
    # Create a new GBD object which operates on the given databases
    def __init__(self, db_string, context='cnf', jobs=1, timeout=5000, separator=" ", join_type="LEFT", verbose=False):
        self.databases = db_string.split(os.pathsep)
        self.context = context
        self.jobs = jobs
        self.mutex = multiprocessing.Lock()
        self.timeout = timeout
        self.separator = separator
        self.join_type = join_type
        self.verbose = verbose
        self.database = Database(self.databases, self.verbose)

    def __enter__(self):
        with ExitStack() as stack:
            stack.enter_context(self.database)
            self._stack = stack.pop_all()
        return self

    def __exit__(self, exc_type, exc, traceback):
        self._stack.__exit__(exc_type, exc, traceback)

    def get_databases(self):
        return list(self.database.get_databases())

    def get_database_path(self, dbname):
        return self.database.dpath(dbname)

    # Get all features
    def get_features(self, dbname=None):
        return self.database.get_features(tables=True, views=True, database=dbname)

    # Get all material features
    def get_material_features(self, dbname=None):
        return self.database.get_features(tables=True, views=False, database=dbname)

    # Get all virtual features
    def get_virtual_features(self, dbname=None):
        return self.database.get_features(tables=False, views=True, database=dbname)

    # Check for existence of given feature
    def feature_exists(self, name):
        return name in self.get_features()

    # Creates the given feature
    def create_feature(self, name, default_value=None):
        if not self.feature_exists(name):
            self.database.create_feature(name, default_value)
        else:
            raise GBDException("Feature '{}' does already exist".format(name))

    # Removes the given feature
    def delete_feature(self, name):
        if self.feature_exists(name):
            self.database.delete_feature(name)
        else:
            raise GBDException("Feature '{}' does not exist or is virtual".format(name))

    # Rename the given feature
    def rename_feature(self, old_name, new_name):
        if not self.feature_exists(old_name):
            raise GBDException("Feature '{}' does not exist or is virtual".format(old_name))
        elif self.feature_exists(new_name):
            raise GBDException("Feature '{}' does already exist".format(new_name))
        else:
            self.database.rename_feature(old_name, new_name)

    # Retrieve information about a specific feature
    def get_feature_info(self, name):
        return self.database.feature_info(name)

    def set_attributes_locked(self, arg):
        self.mutex.acquire()
        try:
            # create new connection as cursor can not be shared across threads
            with Database(self.databases, self.verbose) as db:
                for attr in arg:
                    name, hashv, value = attr[0], attr[1], attr[2]
                    if not name in db.get_features():
                        db.create_feature(name, "empty")
                    db.set_values(name, value, [hashv])
        finally:
            self.mutex.release()

    # Set the attribute value for the given hashes
    def set_attribute(self, feature, value, query, hashes=[], force=False):
        if not feature in self.get_material_features():
            raise GBDException("Feature '{}' missing or virtual".format(feature))
        hash_list = hashes
        if query:
            hash_list = [hash[0] for hash in self.query_search(query, hashes)]
        try:
            self.database.set_values(feature, value, hash_list)
        except Exception as err:
            raise GBDException(str(err))

    # Remove the attribute value for the given hashes
    def remove_attributes(self, feature, hash_list):
        if not feature in self.get_material_features():
            raise GBDException("Feature '{}' not found".format(feature))
        self.database.delete_hashes(feature, hash_list)

    def query_search(self, gbd_query=None, hashes=[], resolve=[], collapse="GROUP_CONCAT", group_by="hash"):
        try:
            query_builder = GBDQuery(self.database, self.join_type, collapse)
            sql = query_builder.build_query(gbd_query, hashes, resolve or [], group_by or "hash")
            return self.database.query(sql)
        except sqlite3.OperationalError as err:
            raise GBDException("Database Operational Error: {}".format(str(err)))
        except tatsu.exceptions.FailedParse as err:
            raise GBDException("Parser Error: {}".format(str(err)))



# GBD Benchmark Database (GBD)
# Copyright (C) 2021 Markus Iser, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sqlite3
import typing

from gbd_tool.gbd_hash import HASH_VERSION
from gbd_tool.util import eprint, prepend_context, context_from_name, make_alnum_ul
from gbd_tool.schema import Schema, FeatureInfo


class DatabaseException(Exception):
    pass


class Database:
    INMEMORYDB = "file::memory:?cache=shared"

    def __init__(self, path_list, verbose=False):
        self.verbose = verbose
        self.schemas = self.init_schemas(path_list)
        self.features = self.init_features()
        self.connection = sqlite3.connect(Schema.IN_MEMORY_DB, uri=True, timeout=10)
        self.cursor = self.connection.cursor()
        self.maindb = None
        schema: Schema
        for schema in self.schemas.values():
            if not schema.csv:
                self.cursor.execute("ATTACH DATABASE '{}' AS {}".format(schema.path, schema.dbname))
            if not self.maindb:
                self.maindb = schema.dbname  # target of inserts and updates
        
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.connection.commit()
        self.connection.close()


    def init_schemas(self, path_list) -> typing.Dict[str, Schema]:
        result = dict()
        for path in path_list:
            schema: Schema = Schema(path)
            if not schema.dbname in result:
                result[schema.dbname] = schema
            elif schema.csv:
                result[schema.dbname].absorb(schema)
            else:
                raise DatabaseException("Database name collision on " + schema.dbname)
        return result

    def init_features(self) -> typing.Dict[str, FeatureInfo]:
        result = dict()
        schema: Schema
        for schema in self.schemas.values():
            feature: FeatureInfo
            for feature in schema.features:
                if not feature.name in result:
                    result[feature.name] = feature
                elif feature.column == "hash": 
                    if not Schema.is_main_hash_column(result[feature.name]) and Schema.is_main_hash_column(feature):
                        result[feature.name] = feature
                else:
                    eprint("Warning: Feature name collision on {}. Using first occurence in {}.".format(feature.name, feature.database))
        return result


    def query(self, q):
        if self.verbose:
            eprint(q)
        return self.cursor.execute(q).fetchall()

    def execute(self, q):
        if self.verbose:
            eprint(q)
        self.cursor.execute(q)
        self.connection.commit()


    def dexists(self, dbname):
        return dbname in self.schemas.keys()

    def dmain(self, dbname):
        return dbname == self.maindb

    def dpath(self, dbname):
        return self.schemas[dbname].path
        
    def dcontexts(self, dbname):
        return self.schemas[dbname].contexts
        
    def dtables(self, dbname):
        return self.schemas[dbname].tables
        
    def dviews(self, dbname):
        return self.schemas[dbname].views


    def fexists(self, feature):
        return feature in self.features.keys()

    def fexists_or_raise(self, feature):
        if not self.fexists(feature):
            raise DatabaseException("Feature {} does not exists".format(feature))

    def fdatabase(self, feature):
        return self.features[feature].database

    def fcontext(self, feature):
        return self.features[feature].context

    def ftable(self, feature, full=True):
        return self.features[feature].table if not full else "{}.{}".format(self.fdatabase(feature), self.features[feature].table)

    def fcolumn(self, feature):
        return self.features[feature].column

    def fdefault(self, feature):
        return self.features[feature].default

    def fvirtual(self, feature):
        return self.features[feature].virtual


    def get_databases(self):
        return self.schemas.keys()
  
    def get_features(self, tables=True, views=True, database=None):
        result = []
        for (feature, info) in self.features.items():
            if not views and info.virtual:
                continue
            if not tables and not info.virtual:
                continue
            if database and database != info.database:
                continue
            result.append(feature)
        return result

    def get_tables(self, context = None, dbname = None):
        tables = list()
        for finfo in self.features.values():
            if (not context or context == finfo.context) and (not dbname or dbname == finfo.database) and not finfo.virtual:
                if not finfo.table in tables:
                    tables.append(finfo.table)
        return tables


    def create_feature(self, name, default_value=None, permissive=False):
        if not permissive:  # internal use can be unchecked, e.g., to create the reserved feature local
            Schema.valid_feature_or_raise(name)
        if self.fexists(name):
            if not permissive:
                raise DatabaseException("Feature {} exists".format(name))
            else:
                return
        context = context_from_name(name)
        if default_value is not None:
            features = prepend_context("features", context)
            self.schemas[self.maindb].create_main_feature_table(context_from_name(name))
            self.execute('ALTER TABLE {}.{} ADD {} TEXT NOT NULL DEFAULT {}'.format(self.maindb, features, name, default_value))
            # update schema
            self.features[name] = FeatureInfo(name, self.maindb, context, features, name, default_value, False)
            self.schemas[self.maindb].features.append(self.features[name])
            # initialize hash
            hashfeature = prepend_context("hash", context)
            if not hashfeature in self.features:
                self.features[hashfeature] = FeatureInfo(hashfeature, self.maindb, context, features, "hash", None, False)
                self.schemas[self.maindb].features.append(self.features[hashfeature])
        else:
            self.execute('CREATE TABLE IF NOT EXISTS {}.{} (hash TEXT NOT NULL, value TEXT NOT NULL, CONSTRAINT all_unique UNIQUE(hash, value))'.format(self.maindb, name))
            # insert default values for new hashes into features table
            features = prepend_context("features", context)
            self.execute("CREATE TRIGGER IF NOT EXISTS {}.{}_dval AFTER INSERT ON {} BEGIN INSERT OR IGNORE INTO {} (hash) VALUES (NEW.hash); END".format(self.maindb, name, name, features))
            # create "filename" view for "local" tables
            if name == prepend_context("local", context):
                filename = prepend_context("filename", context)
                self.execute("CREATE VIEW IF NOT EXISTS {}.{} (hash, value) AS SELECT hash, REPLACE(value, RTRIM(value, REPLACE(value, '/', '')), '') FROM {}".format(self.maindb, filename, name))
            # update contexts
            if not context in self.schemas[self.maindb].contexts:
                self.schemas[self.maindb].contexts.append(context)
            # update schema
            self.features[name] = FeatureInfo(name, self.maindb, context, name, "value", None, False)
            self.schemas[self.maindb].features.append(self.features[name])
            self.schemas[self.maindb].tables.append(name)
            # initialize hash
            hashfeature = prepend_context("hash", context)
            if not hashfeature in self.features:
                self.features[hashfeature] = FeatureInfo(hashfeature, self.maindb, context, name, "hash", None, False)
                self.schemas[self.maindb].features.append(self.features[hashfeature])


    def rename_feature(self, old_name, new_name):
        Schema.valid_feature_or_raise(new_name)
        self.fexists_or_raise(old_name)
        if not self.features[old_name].default:
            self.execute("ALTER TABLE {} RENAME TO {}".format(old_name, new_name))
        else:
            table = self.features[old_name].table
            self.execute("ALTER TABLE {} RENAME COLUMN {} TO {}".format(table, old_name, new_name))
        self.features[new_name] = self.features.pop(old_name)
        self.features[new_name].name = new_name

    def delete_feature(self, name):
        self.fexists_or_raise(name)
        if not self.features[name].default:
            self.execute('DROP TABLE IF EXISTS {}'.format(name))
            self.execute('DROP TRIGGER IF EXISTS {}_dval'.format(name))
        else:
            table = self.features[name].table
            self.execute("ALTER TABLE {} DROP COLUMN {}".format(table, name))
        self.features.pop(name)

    def set_values(self, feature, value, hashes):
        self.fexists_or_raise(feature)
        table = self.features[feature].table
        column = self.features[feature].column
        default = self.features[feature].default
        values = ', '.join(["('{}', '{}')".format(hash, value) for hash in hashes])
        if not default:
            self.execute('INSERT INTO {tab} (hash, {col}) VALUES {vals} ON CONFLICT(hash, value) DO UPDATE SET value=excluded.value'.format(tab=table, col=column, vals=values))
        else:
            self.execute("INSERT INTO {tab} (hash, {col}) VALUES {vals} ON CONFLICT(hash) DO UPDATE SET {col}=excluded.{col}".format(tab=table, col=column, vals=values))

    def delete_hashes(self, feature, hashes):
        self.fexists_or_raise(feature)
        default = self.features[feature].default
        if not default:
            self.execute("DELETE FROM {} WHERE hash IN ('{}')".format(feature, "', '".join(hashes)))
        else:
            self.set_values(feature, default, hashes)

    def delete_values(self, feature, values):
        self.fexists_or_raise(feature)
        table = self.features[feature].table
        column = self.features[feature].column
        default = self.features[feature].default
        if not default:
            self.execute("DELETE FROM {} WHERE value IN ('{}')".format(feature, "', '".join(values)))
        else:
            self.execute("UPDATE {} SET {} = '{}' WHERE value IN ('{}')".format(table, column, default, "', '".join(values)))


    def feature_info(self, feature):
        self.fexists_or_raise(feature)
        result = dict()
        table = self.features[feature].table
        column = self.features[feature].column
        default = self.features[feature].default
        result['feature_name'] = feature
        result['feature_count'] = self.query('SELECT COUNT(*) FROM {}'.format(table))[0][0]
        result['feature_default'] = default
        minmax = self.query('SELECT MIN(CAST({col} AS NUMERIC)), MAX(CAST({col} AS NUMERIC)) FROM {tab} WHERE NOT {col} GLOB "*[^0-9.e\-]*" AND {col} LIKE "_%"'.format(col=column, tab=table))
        if len(minmax):
            result['feature_min'] = minmax[0][0]
            result['feature_max'] = minmax[0][1]
        values = self.query('SELECT DISTINCT {col} FROM {tab} WHERE {col} GLOB "*[^0-9.e\-]*" OR {col} NOT LIKE "_%"'.format(col=column, tab=table))
        result['feature_values'] = " ".join([x[0] for x in values])
        return result


# GBD Benchmark Database (GBD)
# Copyright (C) 2021 Jakob Bach and Markus Iser, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gbd_tool.gbd_api import GBD
from gbd_tool.util import is_number

import mip
import pandas as pd

def optimal_comb(api: GBD, query, runtimes, timeout, k):
    result = api.query_search(query, [], runtimes)
    result = [[int(float(val)) if is_number(val) and float(val) < float(timeout) else int(2*timeout) for val in row[1:]] for row in result]
    dataset = pd.DataFrame(result, columns=runtimes)
    dataset = dataset[(dataset != 2*timeout).any(axis='columns')]
    model = mip.Model()
    instance_solver_vars = [[model.add_var(f'x_{i}_{j}', var_type=mip.BINARY)
                            for j in range(dataset.shape[1])] for i in range(dataset.shape[0])]
    solver_vars = [model.add_var(f's_{j}', var_type=mip.BINARY)for j in range(dataset.shape[1])]
    for var_list in instance_solver_vars:  # per-instance constraints
        model.add_constr(mip.xsum(var_list) == 1)
    for j in range(dataset.shape[1]):  # per-solver-constraints
        model.add_constr(mip.xsum(instance_solver_vars[i][j] for i in range(dataset.shape[0])) <=
                        dataset.shape[0] * solver_vars[j])  # "Implies" in Z3
    model.add_constr(mip.xsum(solver_vars) <= k)
    model.objective = mip.minimize(mip.xsum(instance_solver_vars[i][j] * int(dataset.iloc[i, j])
                                            for i in range(dataset.shape[0]) for j in range(dataset.shape[1])))
    print(dataset.sum().min())
    print(model.optimize())
    print(model.objective_value)
    for index, item in enumerate([var.x for var in solver_vars]):
        if item > 0:
            print(runtimes[index])

    


# GBD Benchmark Database (GBD)
# Copyright (C) 2021 Markus Iser, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



from os.path import isfile, basename

from functools import reduce

import glob

import hashlib

import multiprocessing
import pebble
from concurrent.futures import as_completed

from gbd_tool import config, util
from gbd_tool.gbd_api import GBD, GBDException
from gbd_tool.gbd_hash import gbd_hash
from gbd_tool.util import eprint, confirm, open_cnf_file, slice_iterator

METHOD_UNAVAILABLE = "Method '{}' not available. Install gdbc module from: https://github.com/sat-clique/cnftools"

try:
    from gbdc import extract_base_features
except ImportError:
    def extract_base_features(path, tlim, mlim) -> dict:
        raise GBDException(METHOD_UNAVAILABLE.format("extract_base_features"))

try:
    from gbdc import extract_gate_features
except ImportError:
    def extract_gate_features(path, tlim, mlim) -> dict:
        raise GBDException(METHOD_UNAVAILABLE.format("extract_gate_features"))

try:
    from gbdc import cnf2kis
except ImportError:
    def cnf2kis(in_path, out_path, max_edges, max_nodes) -> dict:
        raise GBDException(METHOD_UNAVAILABLE.format("cnf2kis"))


# Initialize table 'local' with instances found under given path
def init_local(api: GBD, root):
    clocal=util.prepend_context("local", api.context)
    api.database.create_feature(clocal, permissive=True)
    sanitize = [path[0] for path in api.query_search(group_by=clocal) if not isfile(path[0])]
    if len(sanitize) and confirm("{} files not found. Remove stale entries from local table?".format(len(sanitize))):
        for paths in slice_iterator(sanitize, 1000):
            api.database.delete_values("local", paths)
    resultset = []
    for suffix in config.suffix_list(api.context):
        for path in glob.iglob(root + "/**/*" + suffix, recursive=True):
            if not len(api.query_search("{}='{}'".format(clocal, path))):
                resultset.append(("", path))
    run(api, resultset, compute_hash)


def compute_hash(nohashvalue, path, tlim, mlim, args):
    eprint('Hashing {}'.format(path))
    hashvalue = gbd_hash(path)
    return [ ('local', hashvalue, path) ]


# Parallel Runner
def run(api: GBD, resultset, func, tlim = 0, mlim = 0, args=dict()):
    if api.jobs == 1:
        for (hash, local) in resultset:
            api.set_attributes_locked(func(hash, local, tlim, mlim, args))
    else:
        with pebble.ProcessPool(min(multiprocessing.cpu_count(), api.jobs)) as p:
            futures = [ p.schedule(func, (hash, local, tlim, mlim, args)) for (hash, local) in resultset ]
            for f in as_completed(futures, timeout=tlim if tlim > 0 else None):
                try:
                    result = f.result()
                    api.set_attributes_locked(result)
                except pebble.ProcessExpired as e:
                    f.cancel()
                    eprint("{}: {}".format(e.__class__.__name__, e))


def init_transform_cnf_to_kis(api: GBD, query, hashes, tlim, mlim, max_edges, max_nodes):
    api.database.create_feature('kis_local', permissive=True)
    api.database.create_feature('kis_nodes', "empty", permissive=True)
    api.database.create_feature('kis_edges', "empty", permissive=True)
    api.database.create_feature('kis_k', "empty", permissive=True)
    api.database.create_feature('cnf_to_kis', permissive=True)
    api.database.create_feature('kis_to_cnf', permissive=True)
    resultset = api.query_search(query, hashes, ["local"], collapse="MIN")
    run(api, resultset, transform_cnf_to_kis, tlim, mlim, { 'max_edges': max_edges, 'max_nodes': max_nodes })

def transform_cnf_to_kis(cnfhash, cnfpath, tlim, mlim, args):
    if not cnfhash or not cnfpath:
        raise GBDException("Illegal arguments: transform_cnf_to_kis({}, {}, {}, {})".format(cnfhash, cnfpath, tlim, mlim))
    kispath = reduce(lambda path, suffix: path[:-len(suffix)] if path.endswith(suffix) else path, config.suffix_list('cnf'), cnfpath)
    kispath = kispath + ".kis"
    eprint('Transforming {} to k-ISP {}'.format(cnfpath, kispath))
    result = cnf2kis(cnfpath, kispath, args['max_edges'], args['max_nodes'])

    if not "local" in result:
        raise GBDException('''{} (N {}, E {}, K {}) exceeds size limits, cancelled.'''.format(basename(kispath), result['nodes'], result['edges'], result['k']))

    return [ ('kis_local', result['hash'], result['local']),
            ('kis_nodes', result['hash'], result['nodes']), 
            ('kis_edges', result['hash'], result['edges']), 
            ('kis_k', result['hash'], result['k']), 
            ('cnf_to_kis', cnfhash, result['hash']), 
            ('kis_to_cnf', result['hash'], cnfhash) ]


# Initialize base feature tables for given instances
def init_base_features(api: GBD, query, hashes, tlim, mlim):
    resultset = api.query_search(query, hashes, ["local"], collapse="MIN")
    run(api, resultset, base_features, tlim, mlim)

def base_features(hashvalue, filename, tlim, mlim, args):
    eprint('Extracting base features from {}'.format(filename))
    rec = extract_base_features(filename, tlim, mlim)
    eprint('Done with base features from {}'.format(filename))
    return [ (key, hashvalue, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items() ]


# Initialize gate feature tables for given instances
def init_gate_features(api: GBD, query, hashes, tlim, mlim):
    resultset = api.query_search(query, hashes, ["local"], collapse="MIN")
    run(api, resultset, gate_features, tlim, mlim)

def gate_features(hashvalue, filename, tlim, mlim, args):
    eprint('Extracting gate features from {}'.format(filename))
    rec = extract_gate_features(filename, tlim, mlim)
    eprint('Done with gate features from {}'.format(filename))
    return [ (key, hashvalue, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items() ]


# Initialize Graph Features known from Network Analysis
def init_networkit_features(api: GBD, query, hashes, tlim, mlim):
    try:
        import networkit as nk
    except ImportError as e:
        raise GBDException("Module 'networkit' not found. Setup https://networkit.github.io/")
    nk.setNumberOfThreads(min(multiprocessing.cpu_count(), api.jobs))
    resultset = api.query_search(query, hashes, ["local"], collapse="MIN")
    for (hash, local) in resultset: 
        result = networkit_features(hash, local, tlim, mlim)
        eprint(result['hashvalue'])
        for att in result['attributes']:
            eprint(att[1] + "=" + att["2"])

def networkit_features(hashvalue, filename, tlim, mlim):
    rec = dict()
    # TODO: Calculate Networkit Features
    return [ (key, hashvalue, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items() ]


# Initialize degree_sequence_hash for given instances
def init_degree_sequence_hash(api: GBD, hashes, tlim, mlim):
    if not api.feature_exists("degree_sequence_hash"):
        api.create_feature("degree_sequence_hash", "empty")
    resultset = api.query_search(None, hashes, ["local"], collapse="MIN")
    run(api, resultset, compute_degree_sequence_hash)

def compute_degree_sequence_hash(hashvalue, filename, tlim, mlim):
    eprint('Computing degree-sequence hash for {}'.format(filename))
    hash_md5 = hashlib.md5()
    degrees = dict()
    f = open_cnf_file(filename, 'rt')
    for line in f:
        line = line.strip()
        if line and line[0] not in ['p', 'c']:
            for lit in line.split()[:-1]:
                num = int(lit)
                tup = degrees.get(abs(num), (0,0))
                degrees[abs(num)] = (tup[0], tup[1]+1) if num < 0 else (tup[0]+1, tup[1])

    degree_list = list(degrees.values())
    degree_list.sort(key=lambda t: (t[0]+t[1], abs(t[0]-t[1])))
    
    for t in degree_list:
        hash_md5.update(str(t[0]+t[1]).encode('utf-8'))
        hash_md5.update(b' ')
        hash_md5.update(str(abs(t[0]-t[1])).encode('utf-8'))
        hash_md5.update(b' ')

    f.close()

    return [ ('degree_sequence_hash', hashvalue, hash_md5.hexdigest()) ]